@@ -1,19 +1,24 @@
+# Python 3.7 support
+from __future__ import annotations
+
 import argparse
 import csv
 import glob
 import json
 import logging
 import os
+import sys
+import warnings
 from datetime import datetime
 from pathlib import Path
-from typing import List, Dict, Optional, Union
+from typing import List, Dict, Optional, Union, OrderedDict
 
-import sys
 from deprecated import deprecated
 from pygelf import GelfUdpHandler, GelfTcpHandler
 from pytz import utc
 
 from . import dao
+from .dao import ColumnDefinition
 from .exceptions import UserException
 
 
@@ -49,7 +54,8 @@ def init_environment_variables() -> dao.EnvironmentVariables:
                                     url=os.environ.get('KBC_URL', None),
                                     real_user=os.environ.get('KBC_REALUSER', None),
                                     logger_addr=os.environ.get('KBC_LOGGER_ADDR', None),
-                                    logger_port=os.environ.get('KBC_LOGGER_PORT', None)
+                                    logger_port=os.environ.get('KBC_LOGGER_PORT', None),
+                                    data_type_support=os.environ.get('KBC_DATA_TYPE_SUPPORT', None)
                                     )
 
 
@@ -327,7 +333,11 @@ class CommonInterface:
                                  table_metadata: dao.TableMetadata = None,
                                  enclosure: str = '"',
                                  delimiter: str = ',',
-                                 delete_where: dict = None) -> dao.TableDefinition:
+                                 delete_where: dict = None,
+                                 write_always: bool = False,
+                                 schema: Union[
+                                     OrderedDict[str, ColumnDefinition], list[str]] = None,
+                                 has_header: Optional[bool] = None) -> dao.TableDefinition:
         """
                 Helper method for dao.TableDefinition creation along with the "manifest".
                 It initializes path according to the storage_stage type.
@@ -347,6 +357,8 @@ class CommonInterface:
                     enclosure: str: CSV enclosure, by default "
                     delimiter: str: CSV delimiter, by default ,
                     delete_where: Dict with settings for deleting rows
+                    write_always: Bool: If true, the table will be saved to Storage even when the job execution
+                           fails.
         """
         if storage_stage == 'in':
             full_path = os.path.join(self.tables_in_path, name)
@@ -355,6 +367,12 @@ class CommonInterface:
         else:
             raise ValueError(f'Invalid storage_stage value "{storage_stage}". Supported values are: "in" or "out"!')
 
+        # for transition period we need to force legacy mode for csv files w headers.
+        force_legacy_mode = False
+        if not schema and not columns and primary_key:
+            warnings.warn('Primary key is set but columns are not. Forcing legacy mode for CSV file.',
+                          DeprecationWarning)
+            force_legacy_mode = True
         return dao.TableDefinition(name=name,
                                    full_path=full_path,
                                    is_sliced=is_sliced,
@@ -366,7 +384,11 @@ class CommonInterface:
                                    enclosure=enclosure,
                                    delimiter=delimiter,
                                    delete_where=delete_where,
-                                   stage=storage_stage)
+                                   stage=storage_stage,
+                                   write_always=write_always,
+                                   schema=schema,
+                                   has_header=has_header,
+                                   force_legacy_mode=force_legacy_mode)
 
     def create_in_table_definition(self, name: str,
                                    is_sliced: bool = False,
@@ -375,7 +397,8 @@ class CommonInterface:
                                    columns: List[str] = None,
                                    incremental: bool = None,
                                    table_metadata: dao.TableMetadata = None,
-                                   delete_where: str = None) -> dao.TableDefinition:
+                                   delete_where: str = None,
+                                   schema: List[ColumnDefinition] = None) -> dao.TableDefinition:
         """
                        Helper method for input dao.TableDefinition creation along with the "manifest".
                        It initializes path in data/tables/in/ folder.
@@ -389,6 +412,7 @@ class CommonInterface:
                            incremental: Set to true to enable incremental loading
                            table_metadata: <.dao.TableMetadata> object containing column and table metadata
                            delete_where: Dict with settings for deleting rows
+                           schema: Table schema
         """
 
         return self._create_table_definition(name=name,
@@ -399,18 +423,25 @@ class CommonInterface:
                                              columns=columns,
                                              incremental=incremental,
                                              table_metadata=table_metadata,
-                                             delete_where=delete_where)
+                                             delete_where=delete_where,
+                                             schema=schema)
+
+    SCHEMA_TYPE = Union[Dict[str, ColumnDefinition], OrderedDict[str, ColumnDefinition], List[str]]
 
     def create_out_table_definition(self, name: str,
                                     is_sliced: bool = False,
                                     destination: str = '',
                                     primary_key: List[str] = None,
-                                    columns: List[str] = None,
+                                    schema: SCHEMA_TYPE = None,
                                     incremental: bool = None,
                                     table_metadata: dao.TableMetadata = None,
                                     enclosure: str = '"',
                                     delimiter: str = ',',
-                                    delete_where: dict = None) -> dao.TableDefinition:
+                                    delete_where: dict = None,
+                                    write_always: bool = False,
+                                    has_header: Optional[bool] = None,
+                                    **kwargs
+                                    ) -> dao.TableDefinition:
         """
                        Helper method for output dao.TableDefinition creation along with the "manifest".
                        It initializes path in data/tables/out/ folder.
@@ -420,12 +451,18 @@ class CommonInterface:
                            is_sliced: True if the full_path points to a folder with sliced tables
                            destination: String name of the table in Storage.
                            primary_key: List with names of columns used for primary key.
-                           columns: List of columns for headless CSV files
+                           schema: List of columns or mapping of column names and ColumnDefinition objects.
+                            if list of strings is provided, the columns will be created with default settings
+                            (BaseType.string)
                            incremental: Set to true to enable incremental loading
                            table_metadata: <.dao.TableMetadata> object containing column and table metadata
                            enclosure: str: CSV enclosure, by default "
                            delimiter: str: CSV delimiter, by default ,
                            delete_where: Dict with settings for deleting rows
+                           write_always: Bool: If true, the table will be saved to Storage even when the job execution
+                           fails.
+                           has_header:Optional[bool] = flag whether the header is present in the file,
+                                if None legacy method is used
         """
 
         return self._create_table_definition(name=name,
@@ -433,12 +470,15 @@ class CommonInterface:
                                              is_sliced=is_sliced,
                                              destination=destination,
                                              primary_key=primary_key,
-                                             columns=columns,
+                                             columns=kwargs.get('columns'),
                                              incremental=incremental,
                                              table_metadata=table_metadata,
                                              enclosure=enclosure,
                                              delimiter=delimiter,
-                                             delete_where=delete_where)
+                                             delete_where=delete_where,
+                                             write_always=write_always,
+                                             schema=schema,
+                                             has_header=has_header)
 
     # # File processing
 
@@ -862,8 +902,21 @@ class CommonInterface:
     def files_in_path(self):
         return os.path.join(self.data_folder_path, 'in', 'files')
 
-    @staticmethod
-    def write_manifest(io_definition: Union[dao.FileDefinition, dao.TableDefinition]):
+    @property
+    def is_legacy_queue(self) -> bool:
+        """
+        Check if the project is running on legacy queue (v1)
+        Returns:
+
+        """
+        features = os.environ.get('KBC_PROJECT_FEATURE_GATES')
+        is_legacy_queue = True
+        if not features or 'queuev2' in features:
+            is_legacy_queue = False
+        return is_legacy_queue
+
+    def write_manifest(self, io_definition: Union[dao.FileDefinition, dao.TableDefinition],
+                       legacy_manifest: Optional[bool] = None):
         """
         Write a table manifest from dao.IODefinition. Creates the appropriate manifest file in the proper location.
 
@@ -895,30 +948,43 @@ class CommonInterface:
         Returns:
 
         """
-        manifest = io_definition.get_manifest_dictionary()
+
+        if not legacy_manifest:
+
+            legacy_manifest = self._expects_legacy_manifest()
+
+        manifest = io_definition.get_manifest_dictionary(legacy_queue=self.is_legacy_queue,
+                                                         legacy_manifest=legacy_manifest)
         # make dirs if not exist
         os.makedirs(os.path.dirname(io_definition.full_path), exist_ok=True)
         with open(io_definition.full_path + '.manifest', 'w') as manifest_file:
             json.dump(manifest, manifest_file)
 
-    @staticmethod
-    def write_manifests(io_definitions: List[Union[dao.FileDefinition, dao.TableDefinition]]):
+    def _expects_legacy_manifest(self) -> bool:
+        legacy_manifest = self.environment_variables.data_type_support not in ('authoritative', 'hints')
+        om_override = self.configuration.config_data.get('storage', {}).get('output', {}).get('data_type_support')
+        if om_override:
+            legacy_manifest = om_override not in ('authoritative', 'hints')
+        return legacy_manifest
+
+    def write_manifests(self, io_definitions: List[Union[dao.FileDefinition, dao.TableDefinition]],
+                        legacy_manifest: Optional[bool] = None):
         """
         Process all table definition objects and create appropriate manifest files.
         Args:
             io_definitions:
+            legacy_manifest: If True, creates a legacy manifest otherwise, uses the new format if available in project.
 
         Returns:
 
         """
         for io_def in io_definitions:
-            CommonInterface.write_manifest(io_def)
+            self.write_manifest(io_def, legacy_manifest=legacy_manifest)
 
     # ############# DEPRECATED METHODS, TODO: remove
 
-    @staticmethod
     @deprecated(version='1.3.0', reason="You should use write_manifest function")
-    def write_filedef_manifest(file_definition: dao.FileDefinition):
+    def write_filedef_manifest(self, file_definition: dao.FileDefinition):
         """
         Write a table manifest from dao.FileDefinition. Creates the appropriate manifest file in the proper location.
 
@@ -942,11 +1008,10 @@ class CommonInterface:
         Returns:
 
         """
-        CommonInterface.write_manifest(file_definition)
+        self.write_manifest(file_definition)
 
-    @staticmethod
     @deprecated(version='1.3.0', reason="You should use write_manifests function")
-    def write_filedef_manifests(file_definitions: List[dao.FileDefinition]):
+    def write_filedef_manifests(self, file_definitions: List[dao.FileDefinition]):
         """
         Process all table definition objects and create appropriate manifest files.
         Args:
@@ -955,11 +1020,10 @@ class CommonInterface:
         Returns:
 
         """
-        CommonInterface.write_manifests(file_definitions)
+        self.write_manifests(file_definitions)
 
-    @staticmethod
     @deprecated(version='1.3.0', reason="You should use write_manifest function")
-    def write_tabledef_manifest(table_definition: dao.TableDefinition):
+    def write_tabledef_manifest(self, table_definition: dao.TableDefinition):
         """
         Write a table manifest from dao.TableDefinition. Creates the appropriate manifest file in the proper location.
 
@@ -988,11 +1052,10 @@ class CommonInterface:
         Returns:
 
         """
-        CommonInterface.write_manifest(table_definition)
+        self.write_manifest(table_definition, legacy_manifest=True)
 
-    @staticmethod
     @deprecated(version='1.3.0', reason="You should use write_manifests function")
-    def write_tabledef_manifests(table_definitions: List[dao.TableDefinition]):
+    def write_tabledef_manifests(self, table_definitions: List[dao.TableDefinition]):
         """
         Process all table definition objects and create appropriate manifest files.
         Args:
@@ -1001,7 +1064,7 @@ class CommonInterface:
         Returns:
 
         """
-        CommonInterface.write_manifests(table_definitions)
+        self.write_manifests(table_definitions, legacy_manifest=True)
 
 
 # ########## CONFIGURATION
